#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "./heads.h"
#include "./protos.h"

void *Malloc( int, char * );

void matvec(csptr mata, double *x, double *y)
{
/*---------------------------------------------------------------------
| This function does the matrix vector product y = A x.
|----------------------------------------------------------------------
| on entry:
| mata  = the matrix (in SparRow form)
| x     = a vector
|
| on return
| y     = the product A * x
|--------------------------------------------------------------------*/
/*-------------------- local variables                               */
  int i, k, j, *ki;
  double *kr;
  for (i=0; i<mata->n; i++) {
    y[i] = 0.0;
    if((j=mata->nnzrow[i])) {
        kr = mata->ma[i];
        ki = mata->ja[i];
        for (k=0; k<j; k++)
          y[i] += kr[k] * x[ki[k]];
      }
  }
  return;
}
/*---------------end of matvec------------------------------------------
----------------------------------------------------------------------*/

void Lsol(csptr mata, double *b, double *x)
{
/*---------------------------------------------------------------------
| This function does the forward solve L x = b.
| Can be done in place.
|----------------------------------------------------------------------
| on entry:
| mata  = the matrix (in SparRow form)
| b     = a vector
|
| on return
| x     = the solution of L x = b 
|--------------------------------------------------------------------*/
/*   local variables    */
  int i, k;
  double *kr;
  int *ki;
  for (i=0; i<mata->n; i++) {
    x[i] = b[i];
    if ( mata->nnzrow[i] > 0 ) {
      kr = mata->ma[i];
      ki = mata->ja[i];
      for (k=0; k<mata->nnzrow[i]; k++)
	x[i] -= kr[k]*x[ki[k]];
    }
  }
  return;
}
/*---------------end of Lsol-----------------------------------------
----------------------------------------------------------------------*/
void Usol(csptr mata, double *b, double *x)
{
/*---------------------------------------------------------------------
| This function does the backward solve U x = b.
| Can be done in place.
|----------------------------------------------------------------------
| on entry:
| mata  = the matrix (in SparRow form)
| b    = a vector
|
| on return
| x     = the solution of U * x = b 
|
|---------------------------------------------------------------------*/
/*   local variables    */
  int i, k, *ki;
  double *kr;
  for (i=mata->n-1; i>=0; i--) {
    kr = mata->ma[i];
    ki = mata->ja[i];
    x[i] = b[i] ;
    for (k=1; k<mata->nnzrow[i]; k++)
      x[i] -= kr[k] * x[ki[k]];
    x[i] *= kr[0];
  }
  return;
}
/*----------------end of Usol----------------------------------------
----------------------------------------------------------------------*/

int descend(p4ptr levmat, double *x, double *wk)
{
/*---------------------------------------------------------------------
| This function does the (block) forward elimination in ARMS
|                       new       old
|     |            |  |     |    |    |
|     | L        0 |  | wx1  |    | x1 |
|     |            |  |     | =  |    | 
|     | EU^{-1}  I |  | wx2  |    | x2 |
|     |            |  |     |    |    |
| x used and not touched -- or can be the same as wk.
|--------------------------------------------------------------------*/
/*  local variables   */
  int j, len=levmat->n, lenB=levmat->nB, *iperm=levmat->rperm; 
  double *work = levmat->wk; 
/*------------------------------------------------------
|   apply permutation P to rhs 
|-----------------------------------------------------*/
  for (j=0; j<len; j++)
    work[iperm[j]] = x[j] ;
  Lsol(levmat->L, work, wk);       /* sol:   L x = x                 */
  Usol(levmat->U, wk, work);        /* sol:   U work(2) = work         */
/*-------------------- compute x[lenb:.] = x [lenb:.] - E * work(1) */
  matvecz (levmat->E, work, &work[lenB], &wk[lenB]) ; 
  return 0;
}
/*----end-of-descend---------------------------------------------------
|----------------------------------------------------------------------
|--------------------------------------------------------------------*/
int ascend (p4ptr levmat, double *x, double *wk) 
{
/*---------------------------------------------------------------------
| This function does the (block) backward substitution: 
|
|     |            |  |     |    |    |
|     | U  L^{-1}F |  | wk1 |    | x1 |
|     |            |  |     | =  |    |
|     | 0       S  |  | wk2 |    | x2 |  <<-- x2 already computed.
|     |            |  |     |    |    |       and we need x1
|
|    with x2 = S^{-1} wk2 [assumed to have been computed ] 
|--------------------------------------------------------------------*/
 /*--------------------  local variables  */
  int j, len=levmat->n, lenB=levmat->nB, *qperm=levmat->perm;
  double *work = levmat->wk; 
  /*-------------------- copy x onto wk */  
  matvec(levmat->F, &x[lenB], work);   /*  work = F * x_2   */
  Lsol(levmat->L, work, work);         /*  work = L \ work    */
  for (j=0; j<lenB; j++)               /*  wk1 = wk1 - work  */
    work[j] = x[j] - work[j];
  Usol(levmat->U, work, work);         /*  wk1 = U \ wk1 */ 
  memcpy(&work[lenB],&x[lenB],(len-lenB)*sizeof(double));
/*---------------------------------------
|   apply reverse permutation
|--------------------------------------*/
  for (j=0; j<len; j++)
     wk[j] = work[qperm[j]];     
  return 0;
}
/*----end-of-ascend----------------------------------------------------
|----------------------------------------------------------------------
|--------------------------------------------------------------------*/
void matvecz(csptr mata, double *x, double *y, double *z) 
{
/*---------------------------------------------------------------------
| This function does the matrix vector  z = y - A x.
|----------------------------------------------------------------------
| on entry:
| mata  = the matrix (in SparRow form)
| x, y   = two input vector
|
| on return
| z    = the result:  y - A * x
| z-location must be different from that of x 
| i.e., y and x are used but not modified.
|--------------------------------------------------------------------*/
/*   local variables    */
  int i, k,j, *ki;
  double *kr, t;
  for (i=0; i<mata->n; i++) {
    t = y[i] ;
    if((j=mata->nnzrow[i]))
      {
	kr = mata->ma[i];
	ki = mata->ja[i];
	for (k=0; k<j; k++)
	  t -= kr[k] * x[ki[k]];
      }
    z[i] = t; 
  }
  return;
}
/*---------------end of matvecz----------------------------------------
 *--------------------------------------------------------------------*/

